#ident	"$Header: /var/cvsroot/htdb/src/lib/db.c,v 1.79 2004/05/06 23:12:18 del Exp $"

/*
	Copyright  1994-2004 David Whittemore
	<del@adjective.com>, all rights reserved.

	Permission to use, copy, modify, and non-commercially distribute
	this software and its documentation for any purpose is hereby granted
	without fee, provided that the above copyright notice appear in all
	copies and that both that copyright notice and this permission notice
	appear in supporting documentation.  No representations are made about
	the suitability of this software for any purpose.  It is provided
	"as is" without express or implied warranty.
 */
#include	"libhtdb.h"

spaceName	dbSpace = (spaceName)NULL;
int			dbMasterOnly = 0;

#undef	DEBUG

void		dbSync()
{
	dbMasterOnly = 1;
}

void		dbUnsync()
{
	dbMasterOnly = 0;
}

_dbHandle	dbGetSocket(char *handle)
{
	spaceData	*l = db_lookup(handle && *handle ? handle : "master");

	if (l != NULL) {
		htdb_db_t	*data = (htdb_db_t *)l->value;
		return data->socket;
	} else
		return NULL;
}

static char	*dbGetHandle(char **cmd)
{
	char	*pipe = strchr(*cmd, '|'),
			*space = strchr(*cmd, ' ');
	static char	*handle = (char *)NULL;
	if (pipe && *pipe && space && *space && pipe < space) {
		/*
			command was given in "handle|SQL..." format, so we'll use
			the database handle that they've explicitly requested.
			we've checked that the pipe has occurred
			before any spaces, which seems to be a safe assumption
			for legal syntax.
		 */
		*pipe = '\0';
		handle = *cmd;
		*cmd = pipe + 1;
	} else {
		if ( dbMasterOnly )
			return "master";

		/*
			was given in "SQL..." format, so we'll use
			either "master" or "slave" as the database handle
			depending upon SQL context.
		 */
		while ((*cmd) && *(*cmd) && isspace((int)*(*cmd)))
			/*
				eat leading whitespace
			 */
			(*cmd)++;

		/*
			determine which database handle ("master" or "slave") to direct to
			(based upon the simple-minded assumption that "selects"
			will go to the slave, all others go to the master).
		 */
		if (strncasecmp(*cmd, "select", 6) == 0)
			handle = "slave";
		else
			handle = "master";
	}
#ifdef	DEBUG
	{
		FILE	*dp = fopen("/tmp/sqldebug.log", "a+");
		int	col = 0;
		char	*str = *cmd;
		while (str && *str) {
			if (++col > 50 && isspace(*str)) {
				fprintf(dp, "\n\t");
				col = 0;
			}
			fputc(*(str++), dp);
		}
		fprintf(dp, ";\n\n");
		fclose(dp);
	}
#endif
	return handle;
}

static int	htdb_dbQuery(_dbHandle *dbHandle, char *cmd, char **handle)
{
#if	defined(USE_DATABASE)
	*handle = dbGetHandle(&cmd);
	*dbHandle = dbGetSocket(*handle);
	return _dbQuery(*dbHandle, cmd);
#else
	return 0;
#endif
}

static int	htdb_dbBinaryQuery(_dbHandle *dbHandle, char *cmd, unsigned int length)
{
#if	defined(USE_DATABASE)
	*dbHandle = dbGetSocket(dbGetHandle(&cmd));
	return _dbBinaryQuery(*dbHandle, cmd, length);
#else
	return 0;
#endif
}

int	dbSetError(_dbHandle *dbHandle, char *prefix, char *query, int status)
{
#if	defined(USE_DATABASE)
	htdb_setint("_error_database", status);
	htdb_setval("_error_query", query);
	if (htdb_checkval("confDebugLogDBErrors")) {
		char *errstr = (char *)_dbFetchErrorString(*dbHandle);
		if (errstr && *errstr) {
			log_fatal(prefix, query);
			log_fatal(prefix, errstr);
		}
	}
	return status;
#else
	return 0;
#endif
}

_dbResultHandle	*dbSQLQuery(char *cmd, char **handle, int *num_row, int *success)
{
#if	defined(USE_DATABASE)
	_dbHandle	dbHandle;

	*num_row = 0;
	*success = 0;

	if (htdb_dbQuery(&dbHandle, cmd, handle) == 0) {
		*success = 1;
		{
			_dbResultHandle	*resultHandle = mysql_store_result(dbHandle);
			if (resultHandle) {
				*num_row = mysql_num_rows(resultHandle);
				if (*num_row > 0)
					return (resultHandle);
				_dbFreeResultHandle(resultHandle);
			}
		}
	}
#endif
	return NULL;
}

static _dbResultHandle	*dbSQLBinaryQuery(char *cmd, int *num_row, int *success, unsigned int length)
{
#if	defined(USE_DATABASE)
	_dbHandle	dbHandle;

	*num_row = 0;
	*success = 0;

	if (htdb_dbBinaryQuery(&dbHandle, cmd, length) == 0) {
		*success = 1;
		{
			_dbResultHandle	*resultHandle = mysql_store_result(dbHandle);
			if (resultHandle) {
				*num_row = mysql_num_rows(resultHandle);
				if (*num_row > 0)
					return (resultHandle);
				_dbFreeResultHandle(resultHandle);
			}
		}
	}
#endif
	return NULL;
}

int	dbQuery(char *fmt, ...)
{
#if	defined(USE_DATABASE)
	int	found = 1;
	va_list	args;
	char	cmd[BUF_LIMIT_MAX + 1];

	va_start(args, fmt);
	vsnprintf(cmd, BUF_LIMIT_MAX, fmt, args);
	cmd[BUF_LIMIT_MAX] = '\0';
	va_end(args);

	{
		int	success;
		char	*handle;
		_dbResultHandle	*resultHandle = dbSQLQuery(cmd, &handle, &found, &success);
		if (resultHandle)
			_dbFreeResultHandle(resultHandle);
	}
	return found;
#else
	return 0;
#endif
}

int_t	dbInsert(char *fmt, ...)
{
#if	defined(USE_DATABASE)
	int	found = 1;
	va_list	args;
	char	cmd[BUF_LIMIT_MAX + 1];

	va_start(args, fmt);
	vsnprintf(cmd, BUF_LIMIT_MAX, fmt, args);
	cmd[BUF_LIMIT_MAX] = '\0';
	va_end(args);

	{
		int	success = 0;
		char	*handle;
		_dbResultHandle	*resultHandle = dbSQLQuery(cmd, &handle, &found, &success);

		if (resultHandle)
			free(resultHandle);

		if (success)
			return mysql_insert_id(dbGetSocket(handle));
		else {
			_dbHandle dbHandle = dbGetSocket(handle);
			dbSetError(&dbHandle, "insert", (char *) cmd,  ERROR_DATABASE_RESULTHANDLE);
			return 0;
		}
	}
#else
	return 0;
#endif
}

int	dbQueryBinary(char *cmd, unsigned int length)
{
#if	defined(USE_DATABASE)
	int	found = 1;
	{
		int	success;
		_dbResultHandle	*resultHandle = dbSQLBinaryQuery(cmd, &found, &success, length);
		if (resultHandle)
			_dbFreeResultHandle(resultHandle);
		return success;
	}
#else
	return 0;
#endif
}

int	dbDelete(char *fmt, ...)
{
#if	defined(USE_DATABASE)
	int	found = 1;
	va_list	args;
	char	cmd[BUF_LIMIT_MAX + 1];

	va_start(args, fmt);
	vsnprintf(cmd, BUF_LIMIT_MAX, fmt, args);
	cmd[BUF_LIMIT_MAX] = '\0';
	va_end(args);

	{
		int	success;
		char	*handle;
		_dbResultHandle	*resultHandle = dbSQLQuery(cmd, &handle, &found, &success);
		if (resultHandle)
			_dbFreeResultHandle(resultHandle);
	}
	return found;
#else
	return 0;
#endif
}

int	dbUpdate(char *fmt, ...)
{
#if	defined(USE_DATABASE)
	int	success = 0;
	va_list	args;
	char	cmd[BUF_LIMIT_MAX + 1];

	va_start(args, fmt);
	vsnprintf(cmd, BUF_LIMIT_MAX, fmt, args);
	cmd[BUF_LIMIT_MAX] = '\0';
	va_end(args);

	{
		int	found = 0;
		char	*handle;
		_dbResultHandle	*resultHandle = dbSQLQuery(cmd, &handle, &found, &success);

		if (resultHandle) {
			found = 1;
			_dbFreeResultHandle(resultHandle);
		}

	}
	return success;
#else
	return 0;
#endif
}

#define	ISABLOB(x)	!(~((x) | ~(BINARY_FLAG)))

int	dbGetRowByKey(char *table, char *fmt, ...)
{
#if	(defined(USE_MYSQL) && defined(HAVE_LIBMYSQLCLIENT))
	va_list		args;
	char		cmd[BUF_LIMIT_MAX + 1];
	int			rows = 0;
	_dbHandle	dbHandle;
	char		*handle;

	va_start(args, fmt);
	vsnprintf(cmd, BUF_LIMIT_MAX, fmt, args);
	cmd[BUF_LIMIT_MAX] = '\0';
	va_end(args);

#if	defined(USE_DB_RESULT_CACHING)
	if (htdb_getptrint(table, "numResults") > 0 &&
		0 == strcasecmp(htdb_getptrval(table, "query"), cmd)) {
		/*
			whoo-hoo.  cache hit.
			return right now
		 */
		htdb_setptrval(table, "cached", "yes");
		return htdb_getptrint(table, "numResults");
	}
	htdb_setptrval(table, "cached", "no");
	/*
		save the query as the cache key for this result set.
	 */
	htdb_setptrval(table, "query", cmd);
#endif
	/*
		start wid nutin
	 */
	htdb_setptrint(table, "numResults", rows);

	if (htdb_dbQuery(&dbHandle, cmd, &handle) == 0) {
		_dbResultHandle	*resultHandle = mysql_store_result(dbHandle);
		if (resultHandle) {
			_dbResultRow	sqlRow;
			unsigned int	numFields = mysql_num_fields(resultHandle);
			_dbField	*fields = mysql_fetch_fields(resultHandle);
			int	i;

			htdb_setptrint(table, "numFields", numFields);
			for (i = 0 ; i < numFields ; i++) {
				char	*pref = htdb_getptrdyname(table, "field");
				/*
					store field name in:
						prefix->field[]
				 */
				htdb_setarrayval(pref, i + 1, fields[i].name);
				free(pref);
			}

			while ((sqlRow = _dbFetchRow(resultHandle))) {
				unsigned long *lengths = mysql_fetch_lengths(resultHandle);
				for (i = 0 ; i < numFields ; i++) {
					off_t	bytes = lengths[i];
					/*
						array stylee: store as {table}[row]->{field}
					 */
					{
						char	*name = htdb_getobjname(table, rows + 1, fields[i].name),
								*jnk = (ISABLOB(fields[i].flags)) ?
								prepareBinaryFromDatabase(sqlRow[i], &bytes) :
								prepareStringFromDatabase(sqlRow[i]);

						if (jnk && *(jnk)) {
							if (ISABLOB(fields[i].flags))
								htdb_binary(name, jnk, (bytes));
							else
								htdb_store(name, jnk);
						} else
							htdb_setval(name, DATABASE_EMPTY_FIELD);
					}

					/*
						single result/first row stylee: store as '{table}->{field}
					 */
					if (rows == 0) {
						char	name[BUF_LIMIT_SMALL + 1],
								*jnk = (ISABLOB(fields[i].flags)) ?
								prepareBinaryFromDatabase(sqlRow[i], &bytes) :
								prepareStringFromDatabase(sqlRow[i]);
						snprintf(name, BUF_LIMIT_SMALL, "%s->%s",
							table, fields[i].name);
						name[BUF_LIMIT_SMALL] = '\0';
						if (jnk && *(jnk)) {
							if (ISABLOB(fields[i].flags))
								htdb_binary(name, jnk, (bytes));
							else 
								htdb_store(name, jnk);
						} else 
							htdb_setval(name,
								DATABASE_EMPTY_FIELD);
					}
				}
				++rows;
			}
			htdb_setptrint(table, "numResults", rows);
			if (resultHandle)
				_dbFreeResultHandle(resultHandle);
		} else
			return dbSetError(&dbHandle, table,
				cmd, ERROR_DATABASE_RESULTHANDLE);
	} else {
		return dbSetError(&dbHandle, table, cmd, ERROR_DATABASE_QUERY);
	}

	if (rows)
		return rows;

	return dbSetError(&dbHandle, table, cmd, 0);
#else
	return 0;
#endif
}

/*
	Takes two arguments, a query and a function pointer
	that we call back to after each row. 
*/
int	dbCallbackByRow(void (*fptr)(int, int, char **, char **), char *fmt, ...)
{
#if	(defined(USE_MYSQL) && defined(HAVE_LIBMYSQLCLIENT))
	va_list		args;
	char		**fieldBuf;	
	int			rows = 0,
				num_rows;
	_dbHandle	dbHandle;
	char		*handle,
				cmd[BUF_LIMIT_MAX + 1];

	va_start(args, fmt);
	vsnprintf(cmd, BUF_LIMIT_MAX, fmt, args);
	cmd[BUF_LIMIT_MAX] = '\0';
	va_end(args);

	if (htdb_dbQuery(&dbHandle, cmd, &handle) == 0) {
		_dbResultHandle	*resultHandle;
		resultHandle = mysql_store_result(dbHandle);
		if (resultHandle) {
			_dbResultRow	sqlRow;
			unsigned int	numFields = mysql_num_fields(resultHandle);
			_dbField		*fields = mysql_fetch_fields(resultHandle);
			int				i;

			fieldBuf = malloc(sizeof(char *) * numFields);
			for (i = 0; i < numFields; i++) {
				fieldBuf[i] = malloc(sizeof(char) * (strlen(fields[i].name) + 1));
				strcpy(fieldBuf[i], fields[i].name);
			}

			num_rows = mysql_num_rows(resultHandle);

			while ((sqlRow = _dbFetchRow(resultHandle))) {
				++rows;
				fptr(numFields, num_rows, fieldBuf, sqlRow);
			}
			for(i=0; i < numFields; i++)
				free(fieldBuf[i]);
			free(fieldBuf);
			if (resultHandle)
				_dbFreeResultHandle(resultHandle);
		} else
			return dbSetError(&dbHandle, "callbackError", cmd,
				ERROR_DATABASE_RESULTHANDLE);
	} else
		return dbSetError(&dbHandle, "callbackError", cmd, ERROR_DATABASE_QUERY);

	return rows;
#else
	return 0;
#endif
}

int	dbProcessByRow(char *cmd, char *resource_prefix)
{
#if	(defined(USE_MYSQL) && defined(HAVE_LIBMYSQLCLIENT))
	int			rows = 0;
	_dbHandle	dbHandle;
	char		resourceIn[BUF_LIMIT_SMALL + 1],
				resourceOut[BUF_LIMIT_SMALL + 1],
				*handle;

	snprintf(resourceIn, BUF_LIMIT_SMALL, "li%s", resource_prefix);
	resourceIn[BUF_LIMIT_SMALL] = '\0';
	snprintf(resourceOut, BUF_LIMIT_SMALL, "lo%s", resource_prefix);
	resourceOut[BUF_LIMIT_SMALL] = '\0';

	htdb_setptrint(resource_prefix, "numResults", rows);

	if (htdb_dbQuery(&dbHandle, cmd, &handle) == 0) {
		_dbResultHandle	*resultHandle = mysql_store_result(dbHandle);
		if (resultHandle) {
			_dbResultRow	sqlRow;
			unsigned int	numFields = mysql_num_fields(resultHandle);
			_dbField		*fields = mysql_fetch_fields(resultHandle);
			int				i;

			htdb_setval(resourceOut, "");

			htdb_setptrint(resource_prefix, "numFields", numFields);
			for (i = 0 ; i < numFields ; i++) {
				char	*pref = htdb_getptrdyname(resource_prefix, "field");
				/*
					store field name in:
						prefix->field[]
				 */
				htdb_setarrayval(pref, i + 1, fields[i].name);
				free(pref);
			}

			while ((sqlRow = _dbFetchRow(resultHandle))) {
				unsigned long *lengths = mysql_fetch_lengths(resultHandle);
				char	name[BUF_LIMIT_SMALL + 1];

				for (i = 0 ; i < numFields ; i++) {
					off_t	bytes = lengths[i];
					snprintf(name, BUF_LIMIT_SMALL, "%s->loopIndex",
						resource_prefix);
					name[BUF_LIMIT_SMALL] = '\0';
					htdb_setint(name, rows + 1);
					/*
						stylee: {table}->{field}
					 */
					{
						char	*jnk = (ISABLOB(fields[i].flags)) ?
							prepareBinaryFromDatabase(sqlRow[i], &bytes) :
							prepareStringFromDatabase(sqlRow[i]);

						snprintf(name, BUF_LIMIT_SMALL, "%s->%s",
							resource_prefix,
							fields[i].name);
						name[BUF_LIMIT_SMALL] = '\0';
						if (jnk && *(jnk)) {
							if (ISABLOB(fields[i].flags))
								htdb_binary(name, jnk, (bytes));
							else
								htdb_store(name, jnk);
						} else
							htdb_setval(name,
								DATABASE_EMPTY_FIELD);
					}
				}
				htdb_loop(resourceOut, resourceIn);
				for (i = 0 ; i < numFields ; i++) {
					snprintf(name, BUF_LIMIT_SMALL, "%s->%s",
						resource_prefix,
						fields[i].name);
					name[BUF_LIMIT_SMALL] = '\0';
					htdb_setval(name, "");
					snprintf(name, BUF_LIMIT_SMALL, "%s->loopindex",
						resource_prefix);
					name[BUF_LIMIT_SMALL] = '\0';
					htdb_setval(name, "");
				}
				++rows;
			}
			htdb_setptrint(resource_prefix, "numResults", rows);
			if (resultHandle)
				_dbFreeResultHandle(resultHandle);
		} else
			return dbSetError(&dbHandle, resource_prefix, cmd,
				ERROR_DATABASE_RESULTHANDLE);
	} else
		return dbSetError(&dbHandle, resource_prefix, cmd, ERROR_DATABASE_QUERY);

	return rows;
#else
	return 0;
#endif
}

time_t	fetchDatabaseTimestamp(void)
{
#if	defined(USE_DATABASE)
	time_t	out = (time_t)0;
	char	cmd[BUF_LIMIT_SMALL + 1];
	int	found = 0;
	_dbResultHandle	*resultHandle;

	sstrncpy(cmd, "select UNIX_TIMESTAMP()");

	{
		int	success;
		char	*handle;
		resultHandle = dbSQLQuery(cmd, &handle, &found, &success);
	}

	if (found) {
		_dbResultRow	sqlRow;

		sqlRow = _dbFetchRow(resultHandle);

		out = atoi(sqlRow[0]);
	}

	if (resultHandle)
		_dbFreeResultHandle(resultHandle);

	return out;
#else
	return 0;
#endif
}

#if	defined(USE_DBI)

/*
	*** TODO: need libdbi equivalents for: ***

		_dbResultHandle	*dbSQLQuery(char *cmd, char **handle, int *num_row, int *success)
		int	dbQuery(char *fmt, ...)
		int_t	dbInsert(char *fmt, ...)
		int	dbBinaryQuery(char *cmd, unsigned int length)
		int	dbDelete(char *fmt, ...)
		int	dbUpdate(char *fmt, ...)
		int	dbGetRowByKey(char *table, char *fmt, ...)
		int	dbCallbackByRow(void (*fptr)(int, int, char **, char **), char *fmt, ...)
		int	dbProcessByRow(char *cmd, char *resource_prefix)
		time_t	fetchDatabaseTimestamp(void)
 */

#endif

static _dbHandle	dbOpenSocket(char *driver, char *host, char *port, char *dbname, char *username, char *password)
{
	char	*user = (char *)NULL,
			*passwd = (char *)NULL;
	_dbHandle	dbHandle = (_dbHandle)NULL;

	if (username && *username && password && *password) {

		if (*username == '|')
			user = htdb_encryptor("be1", username);
		else
			user = strdup(username);

		if (*password == '|')
			passwd = htdb_encryptor("be1", password);
		else
			passwd = strdup(password);
	}

#if	(defined(USE_MYSQL) && defined(HAVE_LIBMYSQLCLIENT))

	dbHandle = (_dbHandle) malloc(sizeof(MYSQL));

	mysql_init(dbHandle);

	if (username && *username && password && *password) {
		char	*dbhost = (strlen(host) == 0 ||
				strcasecmp(host, "NULL") == 0 ||
				strcasecmp(host, "localhost") == 0) ?
				NULL : host;

		mysql_real_connect(dbHandle, 
			dbhost,
			user,
			passwd, 
			dbname,
			atoi(port), NULL, 0);
	}

	if (dbHandle)
		mysql_select_db(dbHandle, dbname);

#endif

#if	(defined(USE_DBI) && defined(HAVE_LIBDBI))

	dbHandle = dbi_conn_new(driver);

	if (dbHandle) {
		dbi_conn_set_option(dbHandle, "host", host);
		dbi_conn_set_option(dbHandle, "username", user);
		dbi_conn_set_option(dbHandle, "password", passwd);
		dbi_conn_set_option(dbHandle, "dbname", dbname);
		dbi_conn_set_option_numeric(dbHandle "port", port);

		dbi_conn_connect(dbHandle);
	}

#endif

	if (user)
		free(user);
	if (passwd)
		free(passwd);

	return dbHandle;
}

bool_t	doConnect(void)
{
	int	i;
	char	*prefix = "confDatabase";

	/*
		initialize the database handle hash
	 */
	db_initspace();

	/*
		and then loop to find defined database configurations
	 */
	for (i = 0 ; ; i++) {
		char	*name = htdb_getobjval(prefix, i + 1, "name");
		if (name && *name) {
			char	*driver = htdb_getobjval(prefix, i + 1, "driver"),
					*database = htdb_getobjval(prefix, i + 1, "database"),
					*username = htdb_getobjval(prefix, i + 1, "username"),
					*password = htdb_getobjval(prefix, i + 1, "password"),
					*host = htdb_getobjval(prefix, i + 1, "host"),
					*port = htdb_getobjval(prefix, i + 1, "port");
			_dbHandle	socket = dbOpenSocket(driver, host, port, database, username, password);

			if (socket == (_dbHandle)NULL)
				/*
					could not access one of the specified databases
					bomb out completely.
				 */
				return BOOL_FALSE;

			{
				htdb_db_t	*p = (htdb_db_t *)malloc(sizeof(htdb_db_t));
				/*
					populate the dbhandle struct
				 */
				p->name = strdup(name);
				p->driver = strdup(driver);
				p->database = strdup(database);
				p->username = strdup(username);
				p->password = strdup(password);
				p->host = strdup(host);
				p->port = strdup(port);
				p->socket = socket;
				/*
					add this to the database hash
				 */
				db_store(name, p);
			}
		} else
			/*
				no more databases defined
			 */
			break;
	}

	/*
		else return success, since
		we are able to run without any
		valid database.
	 */
	return BOOL_TRUE;
}
